@@ -61,7 +61,7 @@ pub struct FileReference {
     pub access: Option<ReferenceAccess>,
 }
 
-#[derive(Debug, Copy, Clone, PartialEq)]
+#[derive(Debug, Copy, Clone, PartialEq, Eq, Hash)]
 pub enum ReferenceAccess {
     Read,
     Write,
@@ -71,6 +71,7 @@ pub enum ReferenceAccess {
 /// For `pub(crate)` things it's a crate, for `pub` things it's a crate and dependant crates.
 /// In some cases, the location of the references is known to within a `TextRange`,
 /// e.g. for things like local variables.
+#[derive(Clone, Debug)]
 pub struct SearchScope {
     entries: FxHashMap<FileId, Option<TextRange>>,
 }
@@ -215,6 +216,14 @@ impl Definition {
             return SearchScope::crate_graph(db);
         }
 
+        // def is crate root
+        // FIXME: We don't do searches for crates currently, as a crate does not actually have a single name
+        if let &Definition::ModuleDef(hir::ModuleDef::Module(module)) = self {
+            if module.crate_root(db) == module {
+                return SearchScope::reverse_dependencies(db, module.krate());
+            }
+        }
+
         let module = match self.module(db) {
             Some(it) => it,
             None => return SearchScope::empty(),
@@ -223,62 +232,57 @@ impl Definition {
         let file_id = file_id.original_file(db);
 
         if let Definition::Local(var) = self {
-            let range = match var.parent(db) {
-                DefWithBody::Function(f) => f.source(db).map(|src| src.value.syntax().text_range()),
-                DefWithBody::Const(c) => c.source(db).map(|src| src.value.syntax().text_range()),
-                DefWithBody::Static(s) => s.source(db).map(|src| src.value.syntax().text_range()),
+            let def = match var.parent(db) {
+                DefWithBody::Function(f) => f.source(db).map(|src| src.syntax().cloned()),
+                DefWithBody::Const(c) => c.source(db).map(|src| src.syntax().cloned()),
+                DefWithBody::Static(s) => s.source(db).map(|src| src.syntax().cloned()),
             };
-            return match range {
-                Some(range) => SearchScope::file_range(FileRange { file_id, range }),
+            return match def {
+                Some(def) => SearchScope::file_range(def.as_ref().original_file_range(db)),
                 None => SearchScope::single_file(file_id),
             };
         }
 
         if let Definition::SelfType(impl_) = self {
-            return match impl_.source(db).map(|src| src.value.syntax().text_range()) {
-                Some(range) => SearchScope::file_range(FileRange { file_id, range }),
+            return match impl_.source(db).map(|src| src.syntax().cloned()) {
+                Some(def) => SearchScope::file_range(def.as_ref().original_file_range(db)),
                 None => SearchScope::single_file(file_id),
             };
         }
 
         if let Definition::GenericParam(hir::GenericParam::LifetimeParam(param)) = self {
-            let range = match param.parent(db) {
-                hir::GenericDef::Function(it) => {
-                    it.source(db).map(|src| src.value.syntax().text_range())
-                }
-                hir::GenericDef::Adt(it) => {
-                    it.source(db).map(|src| src.value.syntax().text_range())
-                }
-                hir::GenericDef::Trait(it) => {
-                    it.source(db).map(|src| src.value.syntax().text_range())
-                }
-                hir::GenericDef::TypeAlias(it) => {
-                    it.source(db).map(|src| src.value.syntax().text_range())
-                }
-                hir::GenericDef::Impl(it) => {
-                    it.source(db).map(|src| src.value.syntax().text_range())
-                }
-                hir::GenericDef::Variant(it) => {
-                    it.source(db).map(|src| src.value.syntax().text_range())
-                }
-                hir::GenericDef::Const(it) => {
-                    it.source(db).map(|src| src.value.syntax().text_range())
-                }
+            let def = match param.parent(db) {
+                hir::GenericDef::Function(it) => it.source(db).map(|src| src.syntax().cloned()),
+                hir::GenericDef::Adt(it) => it.source(db).map(|src| src.syntax().cloned()),
+                hir::GenericDef::Trait(it) => it.source(db).map(|src| src.syntax().cloned()),
+                hir::GenericDef::TypeAlias(it) => it.source(db).map(|src| src.syntax().cloned()),
+                hir::GenericDef::Impl(it) => it.source(db).map(|src| src.syntax().cloned()),
+                hir::GenericDef::Variant(it) => it.source(db).map(|src| src.syntax().cloned()),
+                hir::GenericDef::Const(it) => it.source(db).map(|src| src.syntax().cloned()),
             };
-            return match range {
-                Some(range) => SearchScope::file_range(FileRange { file_id, range }),
+            return match def {
+                Some(def) => SearchScope::file_range(def.as_ref().original_file_range(db)),
                 None => SearchScope::single_file(file_id),
             };
         }
 
         if let Definition::Macro(macro_def) = self {
-            if macro_def.kind() == hir::MacroKind::Declarative {
-                return if macro_def.attrs(db).by_key("macro_export").exists() {
+            return match macro_def.kind() {
+                hir::MacroKind::Declarative => {
+                    if macro_def.attrs(db).by_key("macro_export").exists() {
+                        SearchScope::reverse_dependencies(db, module.krate())
+                    } else {
+                        SearchScope::krate(db, module.krate())
+                    }
+                }
+                hir::MacroKind::BuiltIn => SearchScope::crate_graph(db),
+                // FIXME: We don't actually see derives in derive attributes as these do not
+                // expand to something that references the derive macro in the output.
+                // We could get around this by emitting dummy `use DeriveMacroPathHere as _;` items maybe?
+                hir::MacroKind::Derive | hir::MacroKind::Attr | hir::MacroKind::ProcMacro => {
                     SearchScope::reverse_dependencies(db, module.krate())
-                } else {
-                    SearchScope::krate(db, module.krate())
-                };
-            }
+                }
+            };
         }
 
         let vis = self.visibility(db);
@@ -311,6 +315,7 @@ impl Definition {
     }
 }
 
+#[derive(Clone)]
 pub struct FindUsages<'a> {
     def: Definition,
     sema: &'a Semantics<'a, RootDatabase>,
@@ -337,7 +342,7 @@ impl<'a> FindUsages<'a> {
         self
     }
 
-    pub fn at_least_one(self) -> bool {
+    pub fn at_least_one(&self) -> bool {
         let mut found = false;
         self.search(&mut |_, _| {
             found = true;
@@ -355,7 +360,7 @@ impl<'a> FindUsages<'a> {
         res
     }
 
-    fn search(self, sink: &mut dyn FnMut(FileId, FileReference) -> bool) {
+    fn search(&self, sink: &mut dyn FnMut(FileId, FileReference) -> bool) {
         let _p = profile::span("FindUsages:search");
         let sema = self.sema;
 
@@ -393,7 +398,7 @@ impl<'a> FindUsages<'a> {
                     continue;
                 }
 
-                if let Some(name) = sema.find_node_at_offset_with_descend(&tree, offset) {
+                for name in sema.find_nodes_at_offset_with_descend(&tree, offset) {
                     if match name {
                         ast::NameLike::NameRef(name_ref) => self.found_name_ref(&name_ref, sink),
                         ast::NameLike::Name(name) => self.found_name(&name, sink),
@@ -410,9 +415,7 @@ impl<'a> FindUsages<'a> {
                         continue;
                     }
 
-                    if let Some(ast::NameLike::NameRef(name_ref)) =
-                        sema.find_node_at_offset_with_descend(&tree, offset)
-                    {
+                    for name_ref in sema.find_nodes_at_offset_with_descend(&tree, offset) {
                         if self.found_self_ty_name_ref(self_ty, &name_ref, sink) {
                             return;
                         }
@@ -444,9 +447,7 @@ impl<'a> FindUsages<'a> {
                         continue;
                     }
 
-                    if let Some(ast::NameLike::NameRef(name_ref)) =
-                        sema.find_node_at_offset_with_descend(&tree, offset)
-                    {
+                    for name_ref in sema.find_nodes_at_offset_with_descend(&tree, offset) {
                         if self.found_self_module_name_ref(&name_ref, sink) {
                             return;
                         }
